@@ -15,8 +15,8 @@ struct TokenizerState<'s> {
     stack: Vec<LexerState>,
     rest: &'s str,
     failed: bool,
-    current_line: usize,
-    current_col: usize,
+    current_line: u32,
+    current_col: u32,
 }
 
 fn find_marker(a: &str) -> Option<(usize, bool)> {
@@ -119,11 +119,11 @@ impl<'s> TokenizerState<'s> {
     }
 
     #[inline(always)]
-    fn loc(&self) -> (usize, usize) {
+    fn loc(&self) -> (u32, u32) {
         (self.current_line, self.current_col)
     }
 
-    fn span(&self, start: (usize, usize)) -> Span {
+    fn span(&self, start: (u32, u32)) -> Span {
         let (start_line, start_col) = start;
         Span {
             start_line,
@@ -139,39 +139,47 @@ impl<'s> TokenizerState<'s> {
     }
 
     fn eat_number(&mut self) -> Result<(Token<'s>, Span), Error> {
+        #[derive(Copy, Clone)]
+        enum State {
+            Integer,      // 123
+            Fraction,     // .123
+            Exponent,     // E | e
+            ExponentSign, // +|-
+        }
+
         let old_loc = self.loc();
-        let mut is_float = false;
-        let num_len = self
+        let mut state = State::Integer;
+        let mut num_len = self
             .rest
-            .as_bytes()
-            .iter()
-            .take_while(|&&c| {
-                if !is_float && c == b'.' {
-                    is_float = true;
-                    true
-                } else {
-                    c.is_ascii_digit()
-                }
-            })
+            .chars()
+            .take_while(|&c| c.is_ascii_digit())
             .count();
+        for c in self.rest.chars().skip(num_len) {
+            state = match (c, state) {
+                ('.', State::Integer) => State::Fraction,
+                ('E' | 'e', State::Integer | State::Fraction) => State::Exponent,
+                ('+' | '-', State::Exponent) => State::ExponentSign,
+                ('0'..='9', State::Exponent) => State::ExponentSign,
+                ('0'..='9', state) => state,
+                _ => break,
+            };
+            num_len += 1;
+        }
+        let is_float = !matches!(state, State::Integer);
+
         let num = self.advance(num_len);
-        Ok(if is_float {
-            (
-                Token::Float(match num.parse::<f64>() {
-                    Ok(val) => val,
-                    Err(_) => return Err(self.syntax_error("invalid float")),
-                }),
-                self.span(old_loc),
-            )
-        } else {
-            (
-                Token::Int(match num.parse::<i64>() {
-                    Ok(val) => val,
-                    Err(_) => return Err(self.syntax_error("invalid integer")),
-                }),
-                self.span(old_loc),
-            )
-        })
+        Ok((
+            ok!(if is_float {
+                num.parse()
+                    .map(Token::Float)
+                    .map_err(|_| self.syntax_error("invalid float"))
+            } else {
+                num.parse()
+                    .map(Token::Int)
+                    .map_err(|_| self.syntax_error("invalid integer"))
+            }),
+            self.span(old_loc),
+        ))
     }
 
     fn eat_identifier(&mut self) -> Result<(Token<'s>, Span), Error> {
