@@ -226,7 +226,7 @@ impl Error {
     pub(crate) fn set_filename_and_span(&mut self, filename: &str, span: Span) {
         self.repr.name = Some(filename.into());
         self.repr.span = Some(span);
-        self.repr.lineno = span.start_line;
+        self.repr.lineno = span.start_line as usize;
     }
 
     pub(crate) fn new_not_found(name: &str) -> Error {
@@ -315,15 +315,6 @@ impl From<fmt::Error> for Error {
     }
 }
 
-impl serde::ser::Error for Error {
-    fn custom<T>(msg: T) -> Self
-    where
-        T: fmt::Display,
-    {
-        Error::new(ErrorKind::BadSerialization, msg.to_string())
-    }
-}
-
 pub fn attach_basic_debug_info<T>(rv: Result<T, Error>, source: &str) -> Result<T, Error> {
     #[cfg(feature = "debug")]
     {
