@@ -233,7 +233,7 @@ pub fn escape(state: &State, v: Value) -> Result<Value, Error> {
     // of the initial state and if that is also not set it falls back
     // to HTML.
     let auto_escape = match state.auto_escape() {
-        AutoEscape::None => match state.env().get_initial_auto_escape(state.name()) {
+        AutoEscape::None => match state.env().initial_auto_escape(state.name()) {
             AutoEscape::None => AutoEscape::Html,
             other => other,
         },
@@ -261,7 +261,7 @@ mod builtins {
     use std::mem;
 
     #[cfg(test)]
-    use similar_asserts::assert_eq;
+    use {crate::testutils::apply_filter, similar_asserts::assert_eq};
 
     /// Converts a value to uppercase.
     ///
@@ -362,7 +362,7 @@ mod builtins {
     #[cfg_attr(docsrs, doc(cfg(feature = "builtins")))]
     pub fn dictsort(v: Value) -> Result<Value, Error> {
         if v.kind() == ValueKind::Map {
-            let mut rv = Vec::new();
+            let mut rv = Vec::with_capacity(v.len().unwrap_or(0));
             let iter = ok!(v.try_iter());
             for key in iter {
                 let value = v.get_item(&key).unwrap_or(Value::UNDEFINED);
@@ -406,7 +406,7 @@ mod builtins {
     #[cfg_attr(docsrs, doc(cfg(feature = "builtins")))]
     pub fn items(v: Value) -> Result<Value, Error> {
         if v.kind() == ValueKind::Map {
-            let mut rv = Vec::new();
+            let mut rv = Vec::with_capacity(v.len().unwrap_or(0));
             let iter = ok!(v.try_iter());
             for key in iter {
                 let value = v.get_item(&key).unwrap_or(Value::UNDEFINED);
@@ -620,8 +620,8 @@ mod builtins {
 
     /// Returns the smallest item from the list.
     #[cfg_attr(docsrs, doc(cfg(feature = "builtins")))]
-    pub fn min(value: Value) -> Result<Value, Error> {
-        let iter = ok!(value.try_iter().map_err(|err| {
+    pub fn min(state: &State, value: Value) -> Result<Value, Error> {
+        let iter = ok!(state.undefined_behavior().try_iter(value).map_err(|err| {
             Error::new(ErrorKind::InvalidOperation, "cannot convert value to list").with_source(err)
         }));
         Ok(iter
@@ -631,8 +631,8 @@ mod builtins {
 
     /// Returns the largest item from the list.
     #[cfg_attr(docsrs, doc(cfg(feature = "builtins")))]
-    pub fn max(value: Value) -> Result<Value, Error> {
-        let iter = ok!(value.try_iter().map_err(|err| {
+    pub fn max(state: &State, value: Value) -> Result<Value, Error> {
+        let iter = ok!(state.undefined_behavior().try_iter(value).map_err(|err| {
             Error::new(ErrorKind::InvalidOperation, "cannot convert value to list").with_source(err)
         }));
         Ok(iter
@@ -642,8 +642,8 @@ mod builtins {
 
     /// Returns the sorted version of the given list.
     #[cfg_attr(docsrs, doc(cfg(feature = "builtins")))]
-    pub fn sort(value: Value, reverse: Option<bool>) -> Result<Value, Error> {
-        let mut items = ok!(value.try_iter().map_err(|err| {
+    pub fn sort(state: &State, value: Value, reverse: Option<bool>) -> Result<Value, Error> {
+        let mut items = ok!(state.undefined_behavior().try_iter(value).map_err(|err| {
             Error::new(ErrorKind::InvalidOperation, "cannot convert value to list").with_source(err)
         }))
         .collect::<Vec<_>>();
@@ -661,8 +661,8 @@ mod builtins {
     /// string this returns the characters.  If the value is undefined
     /// an empty list is returned.
     #[cfg_attr(docsrs, doc(cfg(feature = "builtins")))]
-    pub fn list(value: Value) -> Result<Value, Error> {
-        let iter = ok!(value.try_iter().map_err(|err| {
+    pub fn list(state: &State, value: Value) -> Result<Value, Error> {
+        let iter = ok!(state.undefined_behavior().try_iter(value).map_err(|err| {
             Error::new(ErrorKind::InvalidOperation, "cannot convert value to list").with_source(err)
         }));
         Ok(Value::from(iter.collect::<Vec<_>>()))
@@ -698,16 +698,21 @@ mod builtins {
     /// If you pass it a second argument it’s used to fill missing values on the
     /// last iteration.
     #[cfg_attr(docsrs, doc(cfg(feature = "builtins")))]
-    pub fn slice(value: Value, count: usize, fill_with: Option<Value>) -> Result<Value, Error> {
+    pub fn slice(
+        state: &State,
+        value: Value,
+        count: usize,
+        fill_with: Option<Value>,
+    ) -> Result<Value, Error> {
         if count == 0 {
             return Err(Error::new(ErrorKind::InvalidOperation, "count cannot be 0"));
         }
-        let items = ok!(value.try_iter_owned()).collect::<Vec<_>>();
+        let items = ok!(state.undefined_behavior().try_iter(value)).collect::<Vec<_>>();
         let len = items.len();
         let items_per_slice = len / count;
         let slices_with_extra = len % count;
         let mut offset = 0;
-        let mut rv = Vec::new();
+        let mut rv = Vec::with_capacity(count);
 
         for slice in 0..count {
             let start = offset + slice * items_per_slice;
@@ -750,11 +755,19 @@ mod builtins {
     /// </table>
     /// ```
     #[cfg_attr(docsrs, doc(cfg(feature = "builtins")))]
-    pub fn batch(value: Value, count: usize, fill_with: Option<Value>) -> Result<Value, Error> {
-        let mut rv = Vec::new();
+    pub fn batch(
+        state: &State,
+        value: Value,
+        count: usize,
+        fill_with: Option<Value>,
+    ) -> Result<Value, Error> {
+        if count == 0 {
+            return Err(Error::new(ErrorKind::InvalidOperation, "count cannot be 0"));
+        }
+        let mut rv = Vec::with_capacity(value.len().unwrap_or(0) / count);
         let mut tmp = Vec::with_capacity(count);
 
-        for item in ok!(value.try_iter_owned()) {
+        for item in ok!(state.undefined_behavior().try_iter(value)) {
             if tmp.len() == count {
                 rv.push(Value::from(mem::replace(
                     &mut tmp,
@@ -934,7 +947,7 @@ mod builtins {
         } else {
             None
         };
-        for value in ok!(value.try_iter()) {
+        for value in ok!(state.undefined_behavior().try_iter(value)) {
             let test_value = if let Some(ref attr) = attr {
                 ok!(value.get_path(attr))
             } else {
@@ -1067,7 +1080,7 @@ mod builtins {
         value: Value,
         args: crate::value::Rest<Value>,
     ) -> Result<Vec<Value>, Error> {
-        let mut rv = Vec::new();
+        let mut rv = Vec::with_capacity(value.len().unwrap_or(0));
 
         // attribute mapping
         if args.last().map_or(false, |x| x.is_kwargs()) {
@@ -1085,7 +1098,7 @@ mod builtins {
                     ));
                 }
                 let default = kwargs.get_attr("default").ok();
-                for value in ok!(value.try_iter()) {
+                for value in ok!(state.undefined_behavior().try_iter(value)) {
                     let sub_val = match attr.as_str() {
                         Some(path) => value.get_path(path),
                         None => value.get_item(&attr),
@@ -1113,7 +1126,7 @@ mod builtins {
             .env
             .get_filter(filter_name)
             .ok_or_else(|| Error::from(ErrorKind::UnknownFilter)));
-        for value in ok!(value.try_iter()) {
+        for value in ok!(state.undefined_behavior().try_iter(value)) {
             let new_args = Some(value.clone())
                 .into_iter()
                 .chain(args.0.iter().skip(1).cloned())
@@ -1129,15 +1142,12 @@ mod builtins {
             Ok(a + b)
         }
 
-        let env = crate::Environment::new();
-        State::with_dummy(&env, |state| {
-            let bx = BoxedFilter::new(test);
-            assert_eq!(
-                bx.apply_to(state, &[Value::from(23), Value::from(42)][..])
-                    .unwrap(),
-                Value::from(65)
-            );
-        });
+        let mut env = crate::Environment::new();
+        env.add_filter("test", test);
+        assert_eq!(
+            apply_filter(&env, "test", &[Value::from(23), Value::from(42)]).unwrap(),
+            Value::from(65)
+        );
     }
 
     #[test]
@@ -1146,23 +1156,22 @@ mod builtins {
             rest.iter().fold(val, |a, b| a + b)
         }
 
-        let env = crate::Environment::new();
-        State::with_dummy(&env, |state| {
-            let bx = BoxedFilter::new(sum);
-            assert_eq!(
-                bx.apply_to(
-                    state,
-                    &[
-                        Value::from(1),
-                        Value::from(2),
-                        Value::from(3),
-                        Value::from(4)
-                    ][..]
-                )
-                .unwrap(),
-                Value::from(1 + 2 + 3 + 4)
-            );
-        });
+        let mut env = crate::Environment::new();
+        env.add_filter("sum", sum);
+        assert_eq!(
+            apply_filter(
+                &env,
+                "sum",
+                &[
+                    Value::from(1),
+                    Value::from(2),
+                    Value::from(3),
+                    Value::from(4)
+                ][..]
+            )
+            .unwrap(),
+            Value::from(1 + 2 + 3 + 4)
+        );
     }
 
     #[test]
@@ -1177,31 +1186,30 @@ mod builtins {
             Ok(sum)
         }
 
-        let env = crate::Environment::new();
-        State::with_dummy(&env, |state| {
-            let bx = BoxedFilter::new(add);
-            assert_eq!(
-                bx.apply_to(state, &[Value::from(23), Value::from(42)][..])
-                    .unwrap(),
-                Value::from(65)
-            );
-            assert_eq!(
-                bx.apply_to(
-                    state,
-                    &[Value::from(23), Value::from(42), Value::UNDEFINED][..]
-                )
-                .unwrap(),
-                Value::from(65)
-            );
-            assert_eq!(
-                bx.apply_to(
-                    state,
-                    &[Value::from(23), Value::from(42), Value::from(1)][..]
-                )
-                .unwrap(),
-                Value::from(66)
-            );
-        });
+        let mut env = crate::Environment::new();
+        env.add_filter("add", add);
+        assert_eq!(
+            apply_filter(&env, "add", &[Value::from(23), Value::from(42)][..]).unwrap(),
+            Value::from(65)
+        );
+        assert_eq!(
+            apply_filter(
+                &env,
+                "add",
+                &[Value::from(23), Value::from(42), Value::UNDEFINED][..]
+            )
+            .unwrap(),
+            Value::from(65)
+        );
+        assert_eq!(
+            apply_filter(
+                &env,
+                "add",
+                &[Value::from(23), Value::from(42), Value::from(1)][..]
+            )
+            .unwrap(),
+            Value::from(66)
+        );
     }
 
     #[test]
@@ -1214,24 +1222,24 @@ mod builtins {
             value.into_iter().sum::<i64>()
         }
 
-        let upper = BoxedFilter::new(upper);
-        let sum = BoxedFilter::new(sum);
+        let mut env = crate::Environment::new();
+        env.add_filter("upper", upper);
+        env.add_filter("sum", sum);
 
-        let env = crate::Environment::new();
-        State::with_dummy(&env, |state| {
-            assert_eq!(
-                upper
-                    .apply_to(state, &[Value::from("Hello World!")])
-                    .unwrap(),
-                Value::from("HELLO WORLD!")
-            );
+        assert_eq!(
+            apply_filter(&env, "upper", &[Value::from("Hello World!")]).unwrap(),
+            Value::from("HELLO WORLD!")
+        );
 
-            assert_eq!(
-                sum.apply_to(state, &[Value::from(vec![Value::from(1), Value::from(2)])])
-                    .unwrap(),
-                Value::from(3)
-            );
-        });
+        assert_eq!(
+            apply_filter(
+                &env,
+                "sum",
+                &[Value::from(vec![Value::from(1), Value::from(2)])]
+            )
+            .unwrap(),
+            Value::from(3)
+        );
     }
 
     #[test]
@@ -1244,20 +1252,17 @@ mod builtins {
             rv
         }
 
-        let connect = BoxedFilter::new(connect);
-
-        let env = crate::Environment::new();
-        State::with_dummy(&env, |state| {
-            assert_eq!(
-                connect
-                    .apply_to(
-                        state,
-                        &[Value::from(vec![Value::from("HELLO"), Value::from(42)])]
-                    )
-                    .unwrap(),
-                Value::from("HELLO42")
-            );
-        });
+        let mut env = crate::Environment::new();
+        env.add_filter("connect", connect);
+        assert_eq!(
+            apply_filter(
+                &env,
+                "connect",
+                &[Value::from(vec![Value::from("HELLO"), Value::from(42)])]
+            )
+            .unwrap(),
+            Value::from("HELLO42")
+        );
     }
 }
 
