@@ -5,6 +5,7 @@ use std::ops::{Deref, DerefMut};
 
 use crate::error::{Error, ErrorKind};
 use crate::key::{Key, StaticKey};
+use crate::utils::UndefinedBehavior;
 use crate::value::{
     Arc, MapType, Object, Packed, SeqObject, StringType, Value, ValueKind, ValueRepr,
 };
@@ -140,10 +141,18 @@ pub trait ArgType<'a> {
 
     #[doc(hidden)]
     fn from_state_and_value(
-        _state: Option<&'a State>,
+        state: Option<&'a State>,
         value: Option<&'a Value>,
     ) -> Result<(Self::Output, usize), Error> {
-        Ok((ok!(Self::from_value(value)), 1))
+        if value.map_or(false, |x| x.is_undefined())
+            && state.map_or(false, |x| {
+                matches!(x.undefined_behavior(), UndefinedBehavior::Strict)
+            })
+        {
+            Err(Error::from(ErrorKind::UndefinedError))
+        } else {
+            Ok((ok!(Self::from_value(value)), 1))
+        }
     }
 
     #[doc(hidden)]
@@ -495,7 +504,10 @@ impl<'a> ArgType<'a> for Cow<'_, str> {
     #[inline(always)]
     fn from_value(value: Option<&'a Value>) -> Result<Cow<'a, str>, Error> {
         match value {
-            Some(value) => Ok(value.to_cowstr()),
+            Some(value) => Ok(match value.0 {
+                ValueRepr::String(ref s, _) => Cow::Borrowed(s.as_str()),
+                _ => Cow::Owned(value.to_string()),
+            }),
             None => Err(Error::from(ErrorKind::MissingArgument)),
         }
     }
@@ -578,6 +590,13 @@ impl<'a, T: ArgType<'a, Output = T>> ArgType<'a> for Rest<T> {
 impl<'a> ArgType<'a> for Value {
     type Output = Self;
 
+    fn from_state_and_value(
+        _state: Option<&'a State>,
+        value: Option<&'a Value>,
+    ) -> Result<(Self::Output, usize), Error> {
+        Ok((ok!(Self::from_value(value)), 1))
+    }
+
     fn from_value(value: Option<&'a Value>) -> Result<Self, Error> {
         match value {
             Some(value) => Ok(value.clone()),
