@@ -101,14 +101,12 @@
 // this module is based on the content module in insta which in turn is based
 // on the content module in serde::private::ser.
 
-use std::borrow::Cow;
-use std::cell::RefCell;
+use std::cell::{Cell, RefCell};
 use std::cmp::Ordering;
 use std::collections::BTreeMap;
 use std::convert::TryFrom;
 use std::fmt;
 use std::marker::PhantomData;
-use std::sync::atomic::{self, AtomicBool, AtomicUsize};
 use std::sync::Arc;
 
 use serde::ser::{Serialize, Serializer};
@@ -118,7 +116,7 @@ use crate::functions;
 use crate::key::{Key, StaticKey};
 use crate::utils::OnDrop;
 use crate::value::object::{SimpleSeqObject, SimpleStructObject};
-use crate::value::serialize::ValueSerializer;
+use crate::value::serialize::transform;
 use crate::vm::State;
 
 pub use crate::value::argtypes::{from_args, ArgType, FunctionArgs, FunctionResult, Rest};
@@ -144,13 +142,26 @@ pub(crate) type ValueMap = indexmap::IndexMap<StaticKey, Value>;
 #[cfg(not(feature = "preserve_order"))]
 pub(crate) type ValueMap = std::collections::BTreeMap<StaticKey, Value>;
 
+#[inline(always)]
+pub(crate) fn value_map_with_capacity(capacity: usize) -> ValueMap {
+    #[cfg(not(feature = "preserve_order"))]
+    {
+        let _ = capacity;
+        ValueMap::new()
+    }
+    #[cfg(feature = "preserve_order")]
+    {
+        ValueMap::with_capacity(capacity)
+    }
+}
+
 thread_local! {
-    static INTERNAL_SERIALIZATION: AtomicBool = AtomicBool::new(false);
+    static INTERNAL_SERIALIZATION: Cell<bool> = Cell::new(false);
 
     // This should be an AtomicU64 but sadly 32bit targets do not necessarily have
     // AtomicU64 available.
-    static LAST_VALUE_HANDLE: AtomicUsize = AtomicUsize::new(0);
-    static VALUE_HANDLES: RefCell<BTreeMap<usize, Value>> = RefCell::new(BTreeMap::new());
+    static LAST_VALUE_HANDLE: Cell<u32> = Cell::new(0);
+    static VALUE_HANDLES: RefCell<BTreeMap<u32, Value>> = RefCell::new(BTreeMap::new());
 }
 
 /// Function that returns true when serialization for [`Value`] is taking place.
@@ -170,36 +181,31 @@ thread_local! {
 /// gets sent there, even at the cost of serializing something that cannot be
 /// deserialized.
 pub fn serializing_for_value() -> bool {
-    INTERNAL_SERIALIZATION.with(|flag| flag.load(atomic::Ordering::Relaxed))
+    INTERNAL_SERIALIZATION.with(|flag| flag.get())
 }
 
-/// Enables a temporary code section within which some value
-/// optimizations are enabled.  Currently this is exclusively
-/// used to automatically intern keys when the `key_interning`
-/// feature is enabled.
+/// When key interning is enabled while the returned token is held keys are
+/// interned in a map, freed at the end of the section.
+#[cfg(feature = "key_interning")]
+pub(crate) use crate::key::key_interning::use_string_cache as value_optimization;
+
+/// Without key interning this is a dummy drop.
+#[cfg(not(feature = "key_interning"))]
 #[inline(always)]
-pub(crate) fn with_value_optimization<R, F: FnOnce() -> R>(f: F) -> R {
-    #[cfg(not(feature = "key_interning"))]
-    {
-        f()
-    }
-    #[cfg(feature = "key_interning")]
-    {
-        crate::key::key_interning::with(f)
-    }
+pub(crate) fn value_optimization() -> impl Drop {
+    OnDrop::new(|| {})
 }
 
-/// Executes code within the context of internal serialization
-/// which causes the value type to enable the internal round
-/// tripping serialization.
-fn with_internal_serialization<R, F: FnOnce() -> R>(f: F) -> R {
-    INTERNAL_SERIALIZATION.with(|flag| {
-        let old = flag.load(atomic::Ordering::Relaxed);
-        flag.store(true, atomic::Ordering::Relaxed);
-        let _on_drop = OnDrop::new(|| {
-            flag.store(old, atomic::Ordering::Relaxed);
-        });
-        with_value_optimization(f)
+fn mark_internal_serialization() -> impl Drop {
+    let old = INTERNAL_SERIALIZATION.with(|flag| {
+        let old = flag.get();
+        flag.set(true);
+        old
+    });
+    OnDrop::new(move || {
+        if !old {
+            INTERNAL_SERIALIZATION.with(|flag| flag.set(false));
+        }
     })
 }
 
@@ -282,6 +288,7 @@ pub(crate) enum ValueRepr {
     F64(f64),
     Char(char),
     None,
+    Invalid(Arc<String>),
     U128(Packed<u128>),
     I128(Packed<i128>),
     String(Arc<String>, StringType),
@@ -301,6 +308,7 @@ impl fmt::Debug for ValueRepr {
             ValueRepr::F64(val) => fmt::Debug::fmt(val, f),
             ValueRepr::Char(val) => fmt::Debug::fmt(val, f),
             ValueRepr::None => write!(f, "None"),
+            ValueRepr::Invalid(ref val) => write!(f, "<invalid value: {}>", val),
             ValueRepr::U128(val) => fmt::Debug::fmt(&{ val.0 }, f),
             ValueRepr::I128(val) => fmt::Debug::fmt(&{ val.0 }, f),
             ValueRepr::String(val, _) => fmt::Debug::fmt(val, f),
@@ -320,13 +328,20 @@ impl PartialEq for Value {
     fn eq(&self, other: &Self) -> bool {
         match (&self.0, &other.0) {
             (ValueRepr::None, ValueRepr::None) => true,
+            (ValueRepr::Undefined, ValueRepr::Undefined) => true,
             (ValueRepr::String(ref a, _), ValueRepr::String(ref b, _)) => a == b,
             (ValueRepr::Bytes(a), ValueRepr::Bytes(b)) => a == b,
             _ => match ops::coerce(self, other) {
                 Some(ops::CoerceResult::F64(a, b)) => a == b,
                 Some(ops::CoerceResult::I128(a, b)) => a == b,
                 Some(ops::CoerceResult::Str(a, b)) => a == b,
-                None => false,
+                None => {
+                    if let (Some(a), Some(b)) = (self.as_seq(), other.as_seq()) {
+                        a.iter().eq(b.iter())
+                    } else {
+                        false
+                    }
+                }
             },
         }
     }
@@ -338,6 +353,7 @@ impl PartialOrd for Value {
     fn partial_cmp(&self, other: &Self) -> Option<Ordering> {
         match (&self.0, &other.0) {
             (ValueRepr::None, ValueRepr::None) => Some(Ordering::Equal),
+            (ValueRepr::Undefined, ValueRepr::Undefined) => Some(Ordering::Equal),
             (ValueRepr::Bytes(a), ValueRepr::Bytes(b)) => a.partial_cmp(b),
             _ => match ops::coerce(self, other) {
                 Some(ops::CoerceResult::F64(a, b)) => a.partial_cmp(&b),
@@ -377,6 +393,7 @@ impl fmt::Display for Value {
             }
             ValueRepr::Char(val) => write!(f, "{val}"),
             ValueRepr::None => write!(f, "none"),
+            ValueRepr::Invalid(ref val) => write!(f, "<invalid value: {}>", val),
             ValueRepr::I128(val) => write!(f, "{}", { val.0 }),
             ValueRepr::String(val, _) => write!(f, "{val}"),
             ValueRepr::Bytes(val) => write!(f, "{}", String::from_utf8_lossy(val)),
@@ -452,8 +469,18 @@ impl Value {
     /// # use minijinja::value::Value;
     /// let val = Value::from_serializable(&vec![1, 2, 3]);
     /// ```
+    ///
+    /// This method does not fail but it might return a value that is not valid.  Such
+    /// values will when operated on fail in the template engine in most situations.
+    /// This for instance can happen if the underlying implementation of [`Serialize`]
+    /// fails.  There are also cases where invalid objects are silently hidden in the
+    /// engine today.  This is for instance the case for when keys are used in hash maps
+    /// that the engine cannot deal with.  Invalid values are considered an implementation
+    /// detail.  There is currently no API to validate a value.
     pub fn from_serializable<T: Serialize>(value: &T) -> Value {
-        with_internal_serialization(|| Serialize::serialize(value, ValueSerializer).unwrap())
+        let _serialization_guard = mark_internal_serialization();
+        let _optimization_guard = value_optimization();
+        transform(value)
     }
 
     /// Creates a value from a safe string.
@@ -570,6 +597,8 @@ impl Value {
             ValueRepr::U128(_) => ValueKind::Number,
             ValueRepr::Seq(_) => ValueKind::Seq,
             ValueRepr::Map(..) => ValueKind::Map,
+            // XXX: invalid values report themselves as maps which is a lie
+            ValueRepr::Invalid(_) => ValueKind::Map,
             ValueRepr::Dynamic(ref dy) => match dy.kind() {
                 // XXX: basic objects should probably not report as map
                 ObjectKind::Plain => ValueKind::Map,
@@ -596,7 +625,7 @@ impl Value {
             ValueRepr::Char(x) => x != '\x00',
             ValueRepr::String(ref x, _) => !x.is_empty(),
             ValueRepr::Bytes(ref x) => !x.is_empty(),
-            ValueRepr::None | ValueRepr::Undefined => false,
+            ValueRepr::None | ValueRepr::Undefined | ValueRepr::Invalid(_) => false,
             ValueRepr::Seq(ref x) => !x.is_empty(),
             ValueRepr::Map(ref x, _) => !x.is_empty(),
             ValueRepr::Dynamic(ref x) => match x.kind() {
@@ -711,21 +740,33 @@ impl Value {
     /// # Ok(()) }
     /// ```
     pub fn get_attr(&self, key: &str) -> Result<Value, Error> {
-        let value = match self.0 {
-            ValueRepr::Map(ref items, _) => {
-                let lookup_key = Key::Str(key);
-                items.get(&lookup_key).cloned()
-            }
+        Ok(match self.0 {
+            ValueRepr::Undefined => return Err(Error::from(ErrorKind::UndefinedError)),
+            ValueRepr::Map(ref items, _) => items.get(&Key::Str(key)).cloned(),
             ValueRepr::Dynamic(ref dy) => match dy.kind() {
-                ObjectKind::Plain | ObjectKind::Seq(_) => None,
                 ObjectKind::Struct(s) => s.get_field(key),
+                ObjectKind::Plain | ObjectKind::Seq(_) => None,
             },
-            ValueRepr::Undefined => {
-                return Err(Error::from(ErrorKind::UndefinedError));
-            }
             _ => None,
-        };
-        Ok(value.unwrap_or(Value::UNDEFINED))
+        }
+        .unwrap_or(Value::UNDEFINED))
+    }
+
+    /// Alternative lookup strategy without error handling exclusively for context
+    /// resolution.
+    ///
+    /// The main difference is that the return value will be `None` if the value is
+    /// unable to look up the key rather than returning `Undefined` and errors will
+    /// also not be created.
+    pub(crate) fn get_attr_fast(&self, key: &str) -> Option<Value> {
+        match self.0 {
+            ValueRepr::Map(ref items, _) => items.get(&Key::Str(key)).cloned(),
+            ValueRepr::Dynamic(ref dy) => match dy.kind() {
+                ObjectKind::Struct(s) => s.get_field(key),
+                ObjectKind::Plain | ObjectKind::Seq(_) => None,
+            },
+            _ => None,
+        }
     }
 
     /// Looks up an index of the value.
@@ -829,7 +870,7 @@ impl Value {
         self.as_object().and_then(|x| x.downcast_ref())
     }
 
-    fn get_item_opt(&self, key: &Value) -> Option<Value> {
+    pub(crate) fn get_item_opt(&self, key: &Value) -> Option<Value> {
         let key = some!(Key::from_borrowed_value(key).ok());
 
         let seq = match self.0 {
@@ -872,15 +913,6 @@ impl Value {
         }
     }
 
-    /// Like `as_str` but always stringifies the value.
-    #[allow(unused)]
-    pub(crate) fn to_cowstr(&self) -> Cow<'_, str> {
-        match &self.0 {
-            ValueRepr::String(ref s, _) => Cow::Borrowed(s.as_str()),
-            _ => Cow::Owned(self.to_string()),
-        }
-    }
-
     /// Calls a method on the value.
     pub(crate) fn call_method(
         &self,
@@ -995,12 +1027,21 @@ impl Serialize for Value {
     {
         // enable round tripping of values
         if serializing_for_value() {
-            use serde::ser::SerializeStruct;
-            let handle = LAST_VALUE_HANDLE.with(|x| x.fetch_add(1, atomic::Ordering::Relaxed));
+            let handle = LAST_VALUE_HANDLE.with(|x| {
+                // we are okay with overflowing the handle here because these values only
+                // live for a very short period of time and it's not likely that you run out
+                // of an entire u32 worth of handles in a single serialization operation.
+                // This lets us stick the handle into a unit variant in the serde data model.
+                let rv = x.get().wrapping_add(1);
+                x.set(rv);
+                rv
+            });
             VALUE_HANDLES.with(|handles| handles.borrow_mut().insert(handle, self.clone()));
-            let mut s = ok!(serializer.serialize_struct(VALUE_HANDLE_MARKER, 1));
-            ok!(s.serialize_field("handle", &handle));
-            return s.end();
+            return serializer.serialize_unit_variant(
+                VALUE_HANDLE_MARKER,
+                handle,
+                VALUE_HANDLE_MARKER,
+            );
         }
 
         match self.0 {
@@ -1009,7 +1050,9 @@ impl Serialize for Value {
             ValueRepr::I64(i) => serializer.serialize_i64(i),
             ValueRepr::F64(f) => serializer.serialize_f64(f),
             ValueRepr::Char(c) => serializer.serialize_char(c),
-            ValueRepr::None | ValueRepr::Undefined => serializer.serialize_unit(),
+            ValueRepr::None | ValueRepr::Undefined | ValueRepr::Invalid(_) => {
+                serializer.serialize_unit()
+            }
             ValueRepr::U128(u) => serializer.serialize_u128(u.0),
             ValueRepr::I128(i) => serializer.serialize_i128(i.0),
             ValueRepr::String(ref s, _) => serializer.serialize_str(s),
@@ -1167,7 +1210,7 @@ impl ValueIteratorState {
 
 #[test]
 fn test_dynamic_object_roundtrip() {
-    use std::sync::atomic::AtomicUsize;
+    use std::sync::atomic::{self, AtomicUsize};
 
     #[derive(Debug)]
     struct X(AtomicUsize);
