@@ -6,6 +6,7 @@ use std::sync::Arc;
 use crate::environment::Environment;
 use crate::error::{Error, ErrorKind};
 use crate::value::{OwnedValueIterator, Value};
+use crate::vm::closure_object::Closure;
 use crate::vm::loop_object::Loop;
 
 type Locals<'env> = BTreeMap<&'env str, Value>;
@@ -31,6 +32,13 @@ pub(crate) struct Frame<'env> {
     pub(crate) locals: Locals<'env>,
     pub(crate) ctx: Value,
     pub(crate) current_loop: Option<LoopState>,
+
+    // normally a frame does not carry a closure, but it can when a macro is
+    // declared.  Once that happens, all writes to the frames locals are also
+    // duplicated into the closure.  Macros declared on that level, then share
+    // the closure object to enclose the parent values.  This emulates the
+    // behavior of closures in Jinja2.
+    pub(crate) closure: Option<Arc<Closure>>,
 }
 
 impl<'env> Default for Frame<'env> {
@@ -45,6 +53,7 @@ impl<'env> Frame<'env> {
             locals: Locals::new(),
             ctx,
             current_loop: None,
+            closure: None,
         }
     }
 }
@@ -69,11 +78,18 @@ impl<'env> fmt::Debug for Frame<'env> {
 }
 
 #[cfg_attr(feature = "internal_debug", derive(Debug))]
-#[derive(Default)]
 pub(crate) struct Stack {
     values: Vec<Value>,
 }
 
+impl Default for Stack {
+    fn default() -> Stack {
+        Stack {
+            values: Vec::with_capacity(16),
+        }
+    }
+}
+
 impl Stack {
     pub fn push(&mut self, arg: Value) {
         self.values.push(arg);
@@ -108,7 +124,6 @@ impl From<Vec<Value>> for Stack {
     }
 }
 
-#[derive(Default)]
 pub(crate) struct Context<'env> {
     stack: Vec<Frame<'env>>,
     outer_stack_depth: usize,
@@ -159,18 +174,70 @@ impl<'env> fmt::Debug for Context<'env> {
     }
 }
 
+impl<'env> Default for Context<'env> {
+    fn default() -> Context<'env> {
+        Context {
+            stack: Vec::with_capacity(32),
+            outer_stack_depth: 0,
+        }
+    }
+}
+
 impl<'env> Context<'env> {
     /// Creates a context
     pub fn new(frame: Frame<'env>) -> Context<'env> {
-        Context {
-            stack: vec![frame],
-            outer_stack_depth: 0,
-        }
+        let mut rv = Context::default();
+        rv.stack.push(frame);
+        rv
     }
 
     /// Stores a variable in the context.
     pub fn store(&mut self, key: &'env str, value: Value) {
-        self.stack.last_mut().unwrap().locals.insert(key, value);
+        let top = self.stack.last_mut().unwrap();
+        if let Some(ref closure) = top.closure {
+            closure.store(key, value.clone());
+        }
+        top.locals.insert(key, value);
+    }
+
+    /// Adds a value to a closure if missing.
+    ///
+    /// All macros declare on a certain level reuse the same closure.  This is done
+    /// to emulate the behavior of how scopes work in Jinja2 in Python.  The
+    /// unfortunate downside is that this has to be done with a `Mutex`.
+    #[cfg(feature = "macros")]
+    pub fn enclose(&mut self, env: &Environment, key: &str) {
+        self.closure()
+            .store_if_missing(key, || self.load(env, key).unwrap_or(Value::UNDEFINED));
+    }
+
+    /// Loads the closure and returns it.
+    #[cfg(feature = "macros")]
+    pub fn closure(&mut self) -> Arc<Closure> {
+        let top = self.stack.last_mut().unwrap();
+        top.closure
+            .get_or_insert_with(|| Arc::new(Closure::default()))
+            .clone()
+    }
+
+    /// Temporarily takes the closure.
+    ///
+    /// This is done because includes are in the same scope as the module that
+    /// triggers the import, but we do not want to allow closures to be modified
+    /// from another file as this would be very confusing.
+    ///
+    /// This means that if you override a variable referenced by a macro after
+    /// including in the parent template, it will not override the value seen by
+    /// the macro.
+    #[cfg(feature = "multi_template")]
+    pub fn take_closure(&mut self) -> Option<Arc<Closure>> {
+        self.stack.last_mut().unwrap().closure.take()
+    }
+
+    /// Puts the closure back.
+    #[cfg(feature = "multi_template")]
+    pub fn reset_closure(&mut self, closure: Option<Arc<Closure>>) {
+        self.stack.last_mut().unwrap().closure = closure;
     }
 
     /// Looks up a variable in the context.
@@ -178,9 +245,7 @@ impl<'env> Context<'env> {
         for frame in self.stack.iter().rev() {
             // look at locals first
             if let Some(value) = frame.locals.get(key) {
-                if !value.is_undefined() {
-                    return Some(value.clone());
-                }
+                return Some(value.clone());
             }
 
             // if we are a loop, check if we are looking up the special loop var.
@@ -190,13 +255,10 @@ impl<'env> Context<'env> {
                 }
             }
 
-            // if the frame context is undefined, we skip the lookup
-            if !frame.ctx.is_undefined() {
-                if let Ok(rv) = frame.ctx.get_attr(key) {
-                    if !rv.is_undefined() {
-                        return Some(rv);
-                    }
-                }
+            // perform a fast lookup.  This one will not produce errors if the
+            // context is undefined or of the wrong type.
+            if let Some(rv) = frame.ctx.get_attr_fast(key) {
+                return Some(rv);
             }
         }
 
@@ -218,7 +280,7 @@ impl<'env> Context<'env> {
 
     /// Returns the current locals.
     #[track_caller]
-    #[cfg(feature = "multi-template")]
+    #[cfg(feature = "multi_template")]
     pub fn current_locals(&mut self) -> &mut Locals<'env> {
         &mut self.stack.last_mut().unwrap().locals
     }
