@@ -9,8 +9,8 @@ use crate::compiler::instructions::{
 use crate::environment::Environment;
 use crate::error::{Error, ErrorKind};
 use crate::output::{CaptureMode, Output};
-use crate::utils::AutoEscape;
-use crate::value::{self, ops, MapType, Value, ValueMap, ValueRepr};
+use crate::utils::{AutoEscape, UndefinedBehavior};
+use crate::value::{self, ops, value_map_with_capacity, MapType, Value, ValueRepr};
 use crate::vm::context::{Context, Frame, LoopState, Stack};
 use crate::vm::loop_object::Loop;
 use crate::vm::state::BlockStack;
@@ -23,6 +23,7 @@ use crate::vm::fuel::FuelTracker;
 
 pub use crate::vm::state::State;
 
+mod closure_object;
 mod context;
 #[cfg(feature = "fuel")]
 mod fuel;
@@ -32,7 +33,7 @@ mod macro_object;
 mod state;
 
 // the cost of a single include against the stack limit.
-#[cfg(feature = "multi-template")]
+#[cfg(feature = "multi_template")]
 const INCLUDE_RECURSION_COST: usize = 10;
 
 // the cost of a single macro call against the stack limit.
@@ -85,61 +86,66 @@ impl<'env> Vm<'env> {
         out: &mut Output,
         auto_escape: AutoEscape,
     ) -> Result<Option<Value>, Error> {
-        value::with_value_optimization(|| {
-            self.eval_state(
-                &mut State {
-                    env: self.env,
-                    ctx: Context::new(Frame::new(root)),
-                    current_block: None,
-                    current_call: None,
-                    auto_escape,
-                    instructions,
-                    blocks: prepare_blocks(blocks),
-                    loaded_templates: BTreeSet::new(),
-                    #[cfg(feature = "macros")]
-                    macros: Arc::new(Vec::new()),
-                    #[cfg(feature = "fuel")]
-                    fuel_tracker: self.env.fuel().map(FuelTracker::new),
-                },
-                out,
-            )
-        })
+        let _guard = value::value_optimization();
+        if let ValueRepr::Invalid(ref err) = root.0 {
+            return Err(Error::new(ErrorKind::BadSerialization, err.to_string()));
+        }
+        self.eval_state(
+            &mut State {
+                env: self.env,
+                ctx: Context::new(Frame::new(root)),
+                current_block: None,
+                current_call: None,
+                auto_escape,
+                instructions,
+                blocks: prepare_blocks(blocks),
+                loaded_templates: BTreeSet::new(),
+                #[cfg(feature = "macros")]
+                macros: Arc::new(Vec::new()),
+                #[cfg(feature = "fuel")]
+                fuel_tracker: self.env.fuel().map(FuelTracker::new),
+            },
+            out,
+        )
     }
 
     /// Evaluate a macro in a state.
     #[cfg(feature = "macros")]
+    #[allow(clippy::too_many_arguments)]
     pub fn eval_macro(
         &self,
         instructions: &Instructions<'env>,
         pc: usize,
-        root: Value,
+        closure: Value,
+        caller: Option<Value>,
         out: &mut Output,
         state: &State,
         args: Vec<Value>,
     ) -> Result<Option<Value>, Error> {
-        value::with_value_optimization(|| {
-            let mut ctx = Context::new(Frame::new(root));
-            ok!(ctx.incr_depth(state.ctx.depth() + MACRO_RECURSION_COST));
-            self.eval_impl(
-                &mut State {
-                    env: self.env,
-                    ctx,
-                    current_block: None,
-                    current_call: None,
-                    auto_escape: state.auto_escape(),
-                    instructions,
-                    blocks: BTreeMap::default(),
-                    loaded_templates: BTreeSet::new(),
-                    #[cfg(feature = "macros")]
-                    macros: state.macros.clone(),
-                    #[cfg(feature = "fuel")]
-                    fuel_tracker: state.fuel_tracker.clone(),
-                },
-                out,
-                Stack::from(args),
-                pc,
-            )
-        })
+        let mut ctx = Context::new(Frame::new(closure));
+        if let Some(caller) = caller {
+            ctx.store("caller", caller);
+        }
+        ok!(ctx.incr_depth(state.ctx.depth() + MACRO_RECURSION_COST));
+        self.eval_impl(
+            &mut State {
+                env: self.env,
+                ctx,
+                current_block: None,
+                current_call: None,
+                auto_escape: state.auto_escape(),
+                instructions,
+                blocks: BTreeMap::default(),
+                loaded_templates: BTreeSet::new(),
+                #[cfg(feature = "macros")]
+                macros: state.macros.clone(),
+                #[cfg(feature = "fuel")]
+                fuel_tracker: state.fuel_tracker.clone(),
+            },
+            out,
+            Stack::from(args),
+            pc,
+        )
     }
 
     /// This is the actual evaluation loop that works with a specific context.
@@ -160,6 +166,7 @@ impl<'env> Vm<'env> {
         mut pc: usize,
     ) -> Result<Option<Value>, Error> {
         let initial_auto_escape = state.auto_escape;
+        let undefined_behavior = state.undefined_behavior();
         let mut auto_escape_stack = vec![];
         let mut next_loop_recursion_jump = None;
         let mut loaded_filters = [None; MAX_LOCALS];
@@ -168,7 +175,7 @@ impl<'env> Vm<'env> {
         // If we are extending we are holding the instructions of the target parent
         // template here.  This is used to detect multiple extends and the evaluation
         // uses these instructions when RenderParent is evaluated.
-        #[cfg(feature = "multi-template")]
+        #[cfg(feature = "multi_template")]
         let mut parent_instructions = None;
 
         macro_rules! recurse_loop {
@@ -229,6 +236,16 @@ impl<'env> Vm<'env> {
                 };
             }
 
+            macro_rules! assert_valid {
+                ($expr:expr) => {{
+                    let val = $expr;
+                    if let ValueRepr::Invalid(ref err) = val.0 {
+                        bail!(Error::new(ErrorKind::BadSerialization, err.to_string()));
+                    }
+                    val
+                }};
+            }
+
             // if the fuel consumption feature is enabled, track the fuel
             // consumption here.
             #[cfg(feature = "fuel")]
@@ -249,29 +266,45 @@ impl<'env> Vm<'env> {
                     state.ctx.store(name, stack.pop());
                 }
                 Instruction::Lookup(name) => {
-                    stack.push(state.ctx.load(self.env, name).unwrap_or(Value::UNDEFINED));
+                    stack.push(assert_valid!(state
+                        .lookup(name)
+                        .unwrap_or(Value::UNDEFINED)));
                 }
                 Instruction::GetAttr(name) => {
                     a = stack.pop();
-                    stack.push(ctx_ok!(a.get_attr(name)));
+                    // This is a common enough operation that it's interesting to consider a fast
+                    // path here.  This is slightly faster than the regular attr lookup because we
+                    // do not need to pass down the error object for the more common success case.
+                    // Only when we cannot look up something, we start to consider the undefined
+                    // special case.
+                    stack.push(match a.get_attr_fast(name) {
+                        Some(value) => assert_valid!(value),
+                        None => ctx_ok!(undefined_behavior.handle_undefined(a.is_undefined())),
+                    });
                 }
                 Instruction::GetItem => {
                     a = stack.pop();
                     b = stack.pop();
-                    stack.push(ctx_ok!(b.get_item(&a)));
+                    stack.push(match b.get_item_opt(&a) {
+                        Some(value) => assert_valid!(value),
+                        None => ctx_ok!(undefined_behavior.handle_undefined(b.is_undefined())),
+                    });
                 }
                 Instruction::Slice => {
                     let step = stack.pop();
                     let stop = stack.pop();
                     b = stack.pop();
                     a = stack.pop();
+                    if a.is_undefined() && matches!(undefined_behavior, UndefinedBehavior::Strict) {
+                        bail!(Error::from(ErrorKind::UndefinedError));
+                    }
                     stack.push(ctx_ok!(ops::slice(a, b, stop, step)));
                 }
                 Instruction::LoadConst(value) => {
                     stack.push(value.clone());
                 }
                 Instruction::BuildMap(pair_count) => {
-                    let mut map = ValueMap::new();
+                    let mut map = value_map_with_capacity(*pair_count);
                     for _ in 0..*pair_count {
                         let value = stack.pop();
                         let key = ctx_ok!(stack.pop().try_into_key());
@@ -280,7 +313,7 @@ impl<'env> Vm<'env> {
                     stack.push(Value(ValueRepr::Map(map.into(), MapType::Normal)))
                 }
                 Instruction::BuildKwargs(pair_count) => {
-                    let mut map = ValueMap::new();
+                    let mut map = value_map_with_capacity(*pair_count);
                     for _ in 0..*pair_count {
                         let value = stack.pop();
                         let key = stack.pop().try_into_key().unwrap();
@@ -372,7 +405,7 @@ impl<'env> Vm<'env> {
                     l.object.idx.fetch_add(1, Ordering::Relaxed);
 
                     let next = {
-                        #[cfg(feature = "adjacent-loop-items")]
+                        #[cfg(feature = "adjacent_loop_items")]
                         {
                             let mut triple = l.object.value_triple.lock().unwrap();
                             triple.0 = triple.1.take();
@@ -380,13 +413,13 @@ impl<'env> Vm<'env> {
                             triple.2 = l.iterator.next();
                             triple.1.clone()
                         }
-                        #[cfg(not(feature = "adjacent-loop-items"))]
+                        #[cfg(not(feature = "adjacent_loop_items"))]
                         {
                             l.iterator.next()
                         }
                     };
                     match next {
-                        Some(item) => stack.push(item),
+                        Some(item) => stack.push(assert_valid!(item)),
                         None => {
                             pc = *jump_target;
                             continue;
@@ -424,7 +457,7 @@ impl<'env> Vm<'env> {
                         stack.pop();
                     }
                 }
-                #[cfg(feature = "multi-template")]
+                #[cfg(feature = "multi_template")]
                 Instruction::CallBlock(name) => {
                     if parent_instructions.is_none() {
                         let old_block = state.current_block;
@@ -514,7 +547,7 @@ impl<'env> Vm<'env> {
                         }
                         // leave the one argument on the stack for the recursion
                         recurse_loop!(true);
-                    } else if let Some(func) = state.ctx.load(self.env, name) {
+                    } else if let Some(func) = state.lookup(name) {
                         let args = stack.slice_top(*arg_count);
                         a = ctx_ok!(func.call(state, args));
                         stack.drop_top(*arg_count);
@@ -579,7 +612,7 @@ impl<'env> Vm<'env> {
                 // However for the common case this is convenient because it
                 // lets you put some imports there and for as long as you do not
                 // create name clashes this works fine.
-                #[cfg(feature = "multi-template")]
+                #[cfg(feature = "multi_template")]
                 Instruction::LoadBlocks => {
                     a = stack.pop();
                     if parent_instructions.is_some() {
@@ -591,7 +624,7 @@ impl<'env> Vm<'env> {
                     parent_instructions = Some(ctx_ok!(self.load_blocks(a, state)));
                     out.begin_capture(CaptureMode::Discard);
                 }
-                #[cfg(feature = "multi-template")]
+                #[cfg(feature = "multi_template")]
                 Instruction::RenderParent => {
                     // when an extends statement appears in a template, the last instruction
                     // in that template is the `RenderParent` opcode.  However there is no
@@ -613,15 +646,15 @@ impl<'env> Vm<'env> {
                     pc = 0;
                     continue;
                 }
-                #[cfg(feature = "multi-template")]
+                #[cfg(feature = "multi_template")]
                 Instruction::Include(ignore_missing) => {
                     a = stack.pop();
                     ctx_ok!(self.perform_include(a, state, out, *ignore_missing));
                 }
-                #[cfg(feature = "multi-template")]
+                #[cfg(feature = "multi_template")]
                 Instruction::ExportLocals => {
                     let locals = state.ctx.current_locals();
-                    let mut module = ValueMap::new();
+                    let mut module = value_map_with_capacity(locals.len());
                     for (key, value) in locals.iter() {
                         module.insert((*key).into(), value.clone());
                     }
@@ -633,6 +666,14 @@ impl<'env> Vm<'env> {
                 }
                 #[cfg(feature = "macros")]
                 Instruction::Return => break,
+                #[cfg(feature = "macros")]
+                Instruction::Enclose(name) => {
+                    state.ctx.enclose(state.env, name);
+                }
+                #[cfg(feature = "macros")]
+                Instruction::GetClosure => {
+                    stack.push(Value::from(state.ctx.closure()));
+                }
             }
             pc += 1;
         }
@@ -640,7 +681,7 @@ impl<'env> Vm<'env> {
         Ok(stack.try_pop())
     }
 
-    #[cfg(feature = "multi-template")]
+    #[cfg(feature = "multi_template")]
     fn perform_include(
         &self,
         name: Value,
@@ -677,8 +718,10 @@ impl<'env> Vm<'env> {
             let old_escape = mem::replace(&mut state.auto_escape, tmpl.initial_auto_escape());
             let old_instructions = mem::replace(&mut state.instructions, tmpl.instructions());
             let old_blocks = mem::replace(&mut state.blocks, prepare_blocks(tmpl.blocks()));
+            let old_closure = state.ctx.take_closure();
             ok!(state.ctx.incr_depth(INCLUDE_RECURSION_COST));
             let rv = self.eval_state(state, out);
+            state.ctx.reset_closure(old_closure);
             state.ctx.decr_depth(INCLUDE_RECURSION_COST);
             state.auto_escape = old_escape;
             state.instructions = old_instructions;
@@ -769,7 +812,7 @@ impl<'env> Vm<'env> {
         }
     }
 
-    #[cfg(feature = "multi-template")]
+    #[cfg(feature = "multi_template")]
     fn load_blocks(
         &self,
         name: Value,
@@ -833,7 +876,7 @@ impl<'env> Vm<'env> {
         current_recursion_jump: Option<(usize, bool)>,
     ) -> Result<(), Error> {
         #[allow(unused_mut)]
-        let mut iterator = ok!(iterable.try_iter_owned());
+        let mut iterator = ok!(state.undefined_behavior().try_iter(iterable));
         let len = iterator.len();
         let depth = state
             .ctx
@@ -851,7 +894,7 @@ impl<'env> Vm<'env> {
                     idx: AtomicUsize::new(!0usize),
                     len,
                     depth,
-                    #[cfg(feature = "adjacent-loop-items")]
+                    #[cfg(feature = "adjacent_loop_items")]
                     value_triple: Mutex::new((None, None, iterator.next())),
                     last_changed_value: Mutex::default(),
                 }),
@@ -892,7 +935,7 @@ impl<'env> Vm<'env> {
         name: &str,
         flags: u8,
     ) {
-        use crate::compiler::instructions::{MACRO_CALLER, MACRO_SELF_REFERENTIAL};
+        use crate::compiler::instructions::MACRO_CALLER;
 
         let arg_spec = match stack.pop().0 {
             ValueRepr::Seq(args) => args
@@ -913,7 +956,6 @@ impl<'env> Vm<'env> {
                 arg_spec,
                 macro_ref_id,
                 closure,
-                self_reference: (flags & MACRO_SELF_REFERENTIAL) != 0,
                 caller_reference: (flags & MACRO_CALLER) != 0,
             }),
         }));
